import logger from './logger.js';
import { act, ACTIONS, on, off } from './actions.js';
import { initState } from './state.js';
const channels = new Map();
const stats = {};
const transferRates = {};
export function clients() {
    return Array.from(channels.keys());
}
export function statistics(id) {
    return stats[id];
}
export function createClient(id, channel) {
    channel.onopen = () => {
        logger.debug(id, `data-channel:`, 'open');
        addChannel(id, channel);
    };
    channel.onclose = () => {
        logger.debug(id, `data-channel:`, 'close');
        channel.onerror = channel.onmessage = null;
        removeChannel(id, channel);
    };
    channel.onerror = (error) => {
        var _a;
        if (((_a = error.error) === null || _a === void 0 ? void 0 : _a.message) === 'Transport channel closed')
            return;
        logger.error(id, `data-channel:`, error);
        act(id, ACTIONS.ERROR, error);
    };
    channel.onmessage = msg => {
        const { action, attrs } = JSON.parse(msg.data);
        logger.debug(id, `data-channel:`, action);
        act(id, action, ...(attrs || []));
    };
}
export function destroyClient(id) {
    var _a;
    (_a = channels.get(id)) === null || _a === void 0 ? void 0 : _a.forEach(ch => ch.close());
    channels.delete(id);
    delete stats[id];
    delete transferRates[id];
    act(id, ACTIONS.CLOSE);
    off(id);
    updateClientStates();
}
export function send(id, action, ...attrs) {
    var _a;
    (_a = channels.get(id)) === null || _a === void 0 ? void 0 : _a.forEach(channel => {
        if (channel.readyState === 'open') {
            const msg = JSON.stringify({ action, attrs });
            collectTransferRate(id, msg);
            logger.debug(id, 'send', action);
            try {
                channel.send(msg);
            }
            catch (err) {
                logger.error(id, `could not send to a '${channel.readyState}' channel`, action);
            }
        }
        else {
            logger.debug(id, `could not send to a '${channel.readyState}' channel`, action);
        }
    });
}
export function broadcast(action, ...attrs) {
    for (let id of channels.keys()) {
        send(id, action, ...attrs);
    }
}
export function broadcastToOthers(notThisId, action, ...attrs) {
    for (let id of channels.keys()) {
        if (id !== notThisId)
            send(id, action, ...attrs);
    }
}
function collectTransferRate(id, msg) {
    const collector = transferRates[id] = transferRates[id] || { amount: 0, collectionStarted: Date.now() };
    collector.amount += msg.length;
    const timeCollected = Date.now() - collector.collectionStarted;
    if (timeCollected > 1000) {
        const stat = ensureStatsExist(id);
        stat.dataTransferRate = collector.amount / timeCollected;
        collector.amount = 0;
        collector.collectionStarted = Date.now();
        updateClientStates();
    }
}
function updateLag(id, lag) {
    const stat = ensureStatsExist(id);
    stat.lag = lag;
    updateClientStates();
}
function ensureChannelSetExists(id) {
    const chs = channels.get(id) || new Set();
    channels.set(id, chs);
    return chs;
}
function ensureStatsExist(id) {
    const stat = stats[id] || { lag: Infinity, dataTransferRate: 0 };
    stats[id] = stat;
    return stat;
}
function addChannel(id, channel) {
    ensureStatsExist(id);
    const currentChannels = ensureChannelSetExists(id);
    currentChannels.add(channel);
    for (let ch of currentChannels)
        if (ch !== channel)
            ch.close();
    initState(id);
    updateClientStates();
    act(id, ACTIONS.OPEN);
    on(id, ACTIONS.PING, (theirTime) => updateLag(id, Date.now() - theirTime));
}
function removeChannel(id, channel) {
    const currentChannels = ensureChannelSetExists(id);
    currentChannels.delete(channel);
    if (currentChannels.size === 0)
        destroyClient(id);
}
function updateClientStates() {
    broadcast(ACTIONS.CLIENT_UPDATE, {
        clients: clients(),
        statistics: stats
    });
}
